//***************************************************************************
//
//   TChartUndoRedo component for TeeChartPro VCL
//         version 1.02
//      (c)2004 Ivo Ungermann
//   http://www.dcentral.name/delphi/components
//   mailto://projects@dcentral.name
//
//   Licensed according to GNU GPL version 2
//     http://www.gnu.org/licenses/gpl.html
//
//***************************************************************************


unit ChartUndoRedo;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ActnList, Chart;

type
  TURWatchEvent = (urZoom, urUndoZoom, urScroll);

  TURWatchEvents = set of TURWatchEvent;

type
  TChartUndoRedo = class(TComponent)
  private
    { Private declarations }
    ScrollTimer: TTimer;
    UndoList: TList;
    FUndoRedoPos: Integer;
    FUndoAction, FRedoAction: TAction;
    dummyChart: TChart;
    FChart: TChart;
    FOnZoom: TNotifyEvent;
    FOnUndoZoom: TNotifyEvent;
    FOnScroll: TNotifyEvent;
    FOnUndoActionExecute: TNotifyEvent;
    FOnRedoActionExecute: TNotifyEvent;
    FURWatchEvents: TURWatchEvents;
    FIncludeData: Boolean;
    FScrollTimeout: Integer;
    FUndoBufferDepth: Integer;
    FGroupScroll: Boolean;
    LastWasScroll: Boolean;
    FOnSaveUndoPos: TNotifyEvent;
    FOnUndoRedo: TNotifyEvent;
    function GetUndoRedoMax: Integer;
    procedure OnScrollTimeout(Sender: TObject);
    procedure UROnZoom(Sender: TObject);
    procedure UROnUndoZoom(Sender: TObject);
    procedure UROnScroll(Sender: TObject);
    procedure UROnUndoActionExecute(Sender: TObject);
    procedure UROnRedoActionExecute(Sender: TObject);
    procedure DoStep;
    procedure UpdateURActions;
    procedure SetChart(NewValue: TChart);
    procedure SetUndoAction(NewValue: TAction);
    procedure SetRedoAction(NewValue: TAction);
    procedure SetIncludeData(NewValue: Boolean);
    procedure SetGroupScroll(NewValue: Boolean);
    function GetUndoLevels: Integer;
    procedure SetUndoLevels(NewValue: Integer);
    procedure SetScrollTimeout(NewValue: Integer);
  protected
    { Protected declarations }
  public
    { Public declarations }
    procedure SaveUndoPos;
    function CanUndo: Boolean;
    function CanRedo: Boolean;
    procedure UndoStep;
    procedure RedoStep;
    procedure ClearUndoBuffer;
    property UndoRedoPos: Integer read FUndoRedoPos;
    property UndoRedoMax: Integer read GetUndoRedoMax;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property ScrollCheckInterval: Integer read FScrollTimeout write SetScrollTimeout;
    property UndoLevels: Integer read GetUndoLevels write SetUndoLevels;
    property IncludeData: Boolean read FIncludeData write SetIncludeData;
    property GroupScroll: Boolean read FGroupScroll write SetGroupScroll;
    property UndoAction: TAction read FUndoAction write SetUndoAction;
    property RedoAction: TAction read FRedoAction write SetRedoAction;
    property Chart: TChart read FChart write SetChart;
    property WatchEvents: TURWatchEvents read FURWatchEvents write FURWatchEvents;
    property OnSaveUndoPos: TNotifyEvent read FOnSaveUndoPos write FOnSaveUndoPos;
    property OnUndoRedo: TNotifyEvent read FOnUndoRedo write FOnUndoRedo;
  end;

procedure Register;

implementation

uses TeeEdit, TeeStore;

constructor TChartUndoRedo.Create(AOwner: TComponent);
begin
inherited;
FChart := nil;
FUndoAction := nil;
FRedoAction := nil;
UndoList := tlist.Create;
dummyChart := tchart.Create(nil);
ScrollTimer := TTimer.Create(nil);
ScrollTimer.Enabled := False;
ScrollTimer.OnTimer := OnScrollTimeout;
FUndoRedoPos := -1;
FUndoBufferDepth := 0;
FIncludeData := True;
FGroupScroll := True;
FScrollTimeout := 200;
LastWasScroll := False;
FURWatchEvents := [urZoom, urUndoZoom, urScroll];
end;

destructor TChartUndoRedo.Destroy;
begin
ClearUndoBuffer;
ScrollTimer.Free;
dummyChart.Free;
UndoList.Free;
inherited;
end;

procedure TChartUndoRedo.SetScrollTimeout(NewValue: Integer);
begin
if NewValue = FScrollTimeout then Exit;
if NewValue < 100 then Exit;
FScrollTimeout := NewValue;
end;
function TChartUndoRedo.GetUndoLevels: Integer;
begin
if FUndoBufferDepth = 0
  then Result := FUndoBufferDepth
  else Result := FUndoBufferDepth - 1;
end;

procedure TChartUndoRedo.SetUndoLevels(NewValue: Integer);
begin
if NewValue = UndoLevels then Exit;
if NewValue < 0 then Exit;
if (UndoLevels = 0) or ((NewValue > 0) and (NewValue < UndoLevels)) then
  ClearUndoBuffer;
if NewValue = 0
  then FUndoBufferDepth := NewValue
  else FUndoBufferDepth := NewValue + 1;
end;

function TChartUndoRedo.GetUndoRedoMax: Integer;
begin
Result := UndoList.Count - 1;
end;

procedure TChartUndoRedo.OnScrollTimeout(Sender: TObject);
begin
ScrollTimer.Enabled := False;
if FChart.IPanning.Active then
  begin
  ScrollTimer.Enabled := True;
  Exit;
  end;
if LastWasScroll and FGroupscroll then Dec(FUndoRedoPos);
SaveUndoPos;
LastWasScroll := True;
end;

procedure TChartUndoRedo.UROnScroll(Sender: TObject);
begin
if urScroll in FURWatchEvents then
  begin
  ScrollTimer.Enabled := False;
  ScrollTimer.Interval := FScrollTimeout;
  ScrollTimer.Enabled := True;
  end;
if Assigned(FOnScroll) then
  FOnScroll(Sender);
end;

procedure TChartUndoRedo.SetGroupScroll(NewValue: Boolean);
begin
if NewValue = FGroupScroll then Exit;
FGroupScroll := NewValue;
ClearUndoBuffer;
end;

procedure TChartUndoRedo.SetIncludeData(NewValue: Boolean);
begin
if NewValue = FIncludeData then Exit;
FIncludeData := NewValue;
ClearUndoBuffer;
end;

procedure TChartUndoRedo.SetUndoAction(NewValue: TAction);
begin
if NewValue = FUndoAction then Exit;
if Assigned(FUndoAction) then
  begin
  FUndoAction.OnExecute := FOnUndoActionExecute;
  FOnUndoActionExecute := nil;
  end;
if Assigned(NewValue) then
  begin
  FOnUndoActionExecute := NewValue.OnExecute;
  NewValue.OnExecute := UROnUndoActionExecute;
  end;
FUndoAction := NewValue;
end;

procedure TChartUndoRedo.UROnUndoActionExecute(Sender: TObject);
begin
UndoStep;
if Assigned(FOnUndoActionExecute) then
  FOnUndoActionExecute(Sender);
end;

procedure TChartUndoRedo.SetRedoAction(NewValue: TAction);
begin
if NewValue = FRedoAction then Exit;
if Assigned(FRedoAction) then
  begin
  FRedoAction.OnExecute := FOnRedoActionExecute;
  FOnRedoActionExecute := nil;
  end;
if Assigned(NewValue) then
  begin
  FOnRedoActionExecute := NewValue.OnExecute;
  NewValue.OnExecute := UROnRedoActionExecute;
  end;
FRedoAction := NewValue;
end;

procedure TChartUndoRedo.UROnRedoActionExecute(Sender: TObject);
begin
RedoStep;
if Assigned(FOnRedoActionExecute) then
  FOnRedoActionExecute(Sender);
end;

procedure TChartUndoRedo.SetChart(NewValue: TChart);
begin
if NewValue = FChart then Exit;
if Assigned(FChart) then
  begin
  ClearUndoBuffer;
  FChart.OnZoom := FOnZoom;
  FChart.OnUndoZoom := FOnUndoZoom;
  FChart.OnScroll := FOnScroll;
  FOnZoom := nil;
  FOnUndoZoom := nil;
  FOnScroll := nil;
  end;
if Assigned(NewValue) then
  begin
  FOnZoom := NewValue.OnZoom;
  NewValue.OnZoom := UROnZoom;
  FOnUndoZoom := NewValue.OnUndoZoom;
  NewValue.OnUndoZoom := UROnUndoZoom;
  FOnScroll := NewValue.OnScroll;
  NewValue.OnScroll := UROnScroll;
  end;
FChart := NewValue;
end;

procedure TChartUndoRedo.UROnZoom(Sender: TObject);
begin
if urZoom in FURWatchEvents then
  SaveUndoPos;
if Assigned(FOnZoom) then
  FOnZoom(Sender);
end;

procedure TChartUndoRedo.UROnUndoZoom(Sender: TObject);
begin
if urUndoZoom in FURWatchEvents then
  SaveUndoPos;
if Assigned(FOnUndoZoom) then
  FOnUndoZoom(Sender);
end;

function TChartUndoRedo.CanUndo: Boolean;
begin
Result := FUndoRedoPos > 0;
end;

procedure TChartUndoRedo.UndoStep;
begin
if not Assigned(FChart) then Exit;
if not CanUndo then Exit;
Dec(FUndoRedoPos);
DoStep;
UpdateURActions;
if Assigned(FOnUndoRedo) then
  FOnUndoRedo(Self);
end;

function TChartUndoRedo.CanRedo: Boolean;
begin
Result := (FUndoRedoPos <> -1) and (FUndoRedoPos < UndoList.Count - 1);
end;

procedure TChartUndoRedo.RedoStep;
begin
if not Assigned(FChart) then Exit;
if not CanRedo then Exit;
Inc(FUndoRedoPos);
DoStep;
UpdateURActions;
if Assigned(FOnUndoRedo) then
  FOnUndoRedo(Self);
end;

procedure TChartUndoRedo.DoStep;
begin
FChart.Assign(dummyChart); //!!start from initial state of TChart
TMemoryStream(UndoList[FUndoRedoPos]).Position := 0;
LoadChartFromStream(TCustomChart(FChart), TMemoryStream(UndoList[FUndoRedoPos]));
LastWasScroll := False;
end;

procedure TChartUndoRedo.ClearUndoBuffer;
var
  i: Integer;
begin
for i := 0 to UndoList.Count - 1 do
  TMemoryStream(UndoList[i]).Free;
UndoList.Clear;
FUndoRedoPos := -1;
UpdateURActions;
LastWasScroll := False;
end;

procedure TChartUndoRedo.SaveUndoPos;
var
  ms: TMemoryStream;
  newPos: Boolean;
begin
if not Assigned(FChart) then Exit;
newPos := False;
ms := TMemoryStream.Create;
try
  SaveChartToStream(FChart, ms, FIncludeData);
  if (FUndoRedoPos = -1) or (ms.Size <> TMemoryStream(UndoList[FUndoRedoPos]).Size) or not CompareMem(ms.Memory, TMemoryStream(UndoList[FUndoRedoPos]).Memory, ms.Size) then
    begin
    while UndoList.Count - 1 > FUndoRedoPos do
      begin
      TMemoryStream(UndoList[UndoList.Count - 1]).Free;
      UndoList.Delete(UndoList.Count - 1);
      end;
    if (FUndoBufferDepth > 0) and (FUndoBufferDepth = UndoList.Count) then
      begin
      TMemoryStream(UndoList[0]).Free;
      UndoList.Delete(0);
      Dec(FUndoRedoPos);
      end;
    UndoList.Add(ms);
    Inc(FUndoRedoPos);
    newPos := True;
    end;
  except
  end;
UpdateURActions;
LastWasScroll := False;
if NewPos then
  if Assigned(FOnSaveUndoPos) then
    FOnSaveUndoPos(Self);
end;

procedure TChartUndoRedo.UpdateURActions;
begin
if Assigned(FUndoAction) then FUndoAction.Enabled := CanUndo;
if Assigned(FRedoAction) then FRedoAction.Enabled := CanRedo;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TChartUndoRedo]);
end;

end.
